############################################
# general configuration
############################################

fct.names <- c("constant", "logistp", "logistn", "dlogistp", "dlogistn") # specifying more functions breaks fit(). You should revisit it's code for some function-specific optimizations
lambda <- 1

############################################
# standard library
############################################

# just a helper function that R does not contain somehow
show.me <- function(fct, x=NULL, xlim=par("usr")[1 : 2], length=100, ...) {
  if (is.null(x))
    x <- seq(xlim[1], xlim[2], length.out=length)
  lines(x, apply(as.matrix(x), 1, fct), ...)
}

############################################
# handle input gated data
############################################

read.gated.file <- function(filename) {
  if (substr(filename, nchar(filename) - 2, nchar(filename)) == ".gz") {
    nall <- as.numeric(read.table(gzfile(filename), skip=0, nrows=1))
    nmeta <- as.numeric(read.table(gzfile(filename), skip=1, nrows=1))
    if (nmeta > 0) {
      metadata <- read.table(gzfile(filename), skip=2, nrows=nmeta, sep="\n", blank.lines.skip=FALSE)
    } else {
      metadata <- data.frame()
    }
    n <- as.numeric(read.table(gzfile(filename), skip=(2 + nmeta), nrows=1))
    data <- read.table(gzfile(filename), skip=(2 + nmeta + 1), nrows=n)
  } else {
    nall <- as.numeric(read.table(filename, skip=0, nrows=1))
    nmeta <- as.numeric(read.table(filename, skip=1, nrows=1))
    if (nmeta > 0) {
      metadata <- read.table(filename, skip=2, nrows=nmeta, sep="\n", blank.lines.skip=FALSE)
    } else {
      metadata <- data.frame()
    }
    n <- as.numeric(read.table(filename, skip=(2 + nmeta), nrows=1))
    data <- read.table(filename, skip=(2 + nmeta + 1), nrows=n)
  }
  metadata <- as.matrix(metadata)
  data <- as.matrix(data)
  colnames(data) <- NULL
  rownames(data) <- NULL
  return(list(data=data, metadata=metadata))
}

write.gated.file <- function(filename, metadata, data) {
  write.table(1 + dim(metadata)[1] + 1 + dim(data)[1], filename, append=F, col.names=F, row.names=F, quote=F)
  write.table(dim(metadata)[1],                        filename, append=T, col.names=F, row.names=F, quote=F)
  write.table(metadata,                                filename, append=T, col.names=F, row.names=F, quote=F)
  write.table(dim(data)[1],                            filename, append=T, col.names=F, row.names=F, quote=F)
  write.table(data,                                    filename, append=T, col.names=F, row.names=F, quote=F)
}

# if there are in average min. 5 events with the same measurement time we should
# recalculate the time of the consecutive events as if there were evenly
# distributed between their original time + delta where delta is the smallest
# time difference that exists between arbitrary two events. with this approach
# it is not a problem if there is a big gap between events and there is no
# problem when delta is very small (in this case the time values change very little)
retime <- function(time) {
  u.time <- unique(time)
  delta <- min(u.time[2:length(u.time)]-u.time[1:length(u.time)-1])
  splitted <- split(time, time)
  if (length(time)/length(splitted) > 5) {
    j <- 1
    for (i in 1:length(splitted)) {
      time[j : (j + length(splitted[[i]]) - 1)] <- splitted[[i]][1] + seq(from=0, by=delta/length(splitted[[i]]), length.out=length(splitted[[i]]))
      j <- j + length(splitted[[i]])
    }
  }
  return(time)
}

# this function divides the time range into 100 events (or less, if there is a gap larger than 1/100 of measurement time)
dividetime <- function(times, n=100) {
  step <- (tail(times, 1) - times[1]) / n
  for (i in 1 : n)
    times[which((times < step * i + times[1]) & (times >= step * (i - 1) + times[1]))] <- times[1] + step * (i - 1)
  times[length(times)] <- times[length(times) - 1]
  return(times)
}

# calculates n numbers between 0 and 1 distributed equally with half space in the beginning and end
quantiles <- function(n=20, cutoff=0) {
  i <- 1 - cutoff
  return(seq(cutoff/2+i/n/2, cutoff/2+i-i/n/2, i/n))
}

# input: data: row: one event. 1. column: time. 2. oszlop: intensity value
# output: qdata[i, j] the intensity corresponding to the i^th quantile and j^th time point
quantiles.from.data <- function(data, quantiles=quantiles(20)) {
  qdata <- array(0, dim=c(length(quantiles), length(unique(data[, 1]))))
  splitted <- split(data[, 2], data[, 1])
  for (j in 1 : length(splitted)) #j az aktualis idopont
    qdata[, j] <- quantile(splitted[[j]], quantiles)
  return(qdata)
}



############################################
# functions (kinetic models)
############################################

# constant:
#    y,
#    constraint: y >= 0
constant <- function(x=NULL, p=NULL) {
  # y
  # 1
  if (is.null(p)) return(1) # number of parameters
  return(abs(p))
}

# logistp (logist positive, can only go upwards):
#    y0 + (y2 - y0) / (1 + exp((-x + x1) * 4 * m / (y2 - y0))),
#    constraints: y0, y2, x1, m >= 0 && y0 < y2
logistp <- function(x=NULL, p=NULL) {
  # y0, y2, x1, m
  #  1,  2,  3, 4
  if (is.null(p)) return(4) # number of parameters
  p <- abs(p)
  if (p[1] >= p[2]) p[1:2] <- p[2:1]
  if (is.null(x)) return(p)
  if (p[1] == p[2]) return(x*0 + p[1])
  return(p[1] + (p[2] - p[1]) / (1 + exp((-x + p[3]) * 4 * p[4] / (p[2] - p[1]))))
}

# logistn (logist negative, can only go downwards):
#    y0 + (y2 - y0) / (1 + exp((-x + x1) * 4 * m / (y2 - y0))),
#    constraints: y0, y2, x1 >= 0 && m <= 0 && y2 < y0
logistn <- function(x=NULL, p=NULL) {
  # y0, y2, x1, m
  #  1,  2,  3, 4
  if (is.null(p)) return(4) # number of parameters
  p <- c(abs(p[1:3]), -abs(p[4]))
  if (p[2] >= p[1]) p[1:2] <- p[2:1]
  if (is.null(x)) return(p)
  if (p[1] == p[2]) return(x*0 + p[1])
  return(p[1] + (p[2] - p[1]) / (1 + exp((-x + p[3]) * 4 * p[4] / (p[2] - p[1]))))
}

# dlogistp (dlogist positive, first up, then down):
#    y0 + (y1 - y0) / (1 + ((x1 - x) / xd0) ^ (4 * xd0 * m0 / (y1 - y0))), if x < x1
#    y2 + (y1 - y2) / (1 + ((x - x1) / xd2) ^ (4 * xd2 * m2 / (y2 - y1))), otherwise,
#    constraints: y0, y1, y2, m0, x1, xd0, xd2 >= 0 && m2 <= 0 && xd0 <= x1 && y1 > y0 && y1 > y2
if (Sys.info()["sysname"] != "Linux" && (Sys.info()["machine"] == "x86_32" || Sys.info()["machine"] == "x86-32" || Sys.info()["machine"] == "x86")) {
    dyn.load("dlogistx-32.dll")
} else if (Sys.info()["sysname"] != "Linux" && (Sys.info()["machine"] == "x86_64" || Sys.info()["machine"] == "x86-64")) {
    dyn.load("dlogistx-64.dll")
} else {
    dyn.load("dlogistx.so")
}
dlogistp <- function(x=NULL, p=NULL) {
  # y0, y1, y2, x1, xd0, xd2, m0, m2
  #  1,  2,  3,  4,   5,   6,  7,  8
  if (is.null(p)) return(8)
  if (is.null(x)) {
    p <- abs(p)
    p[8] <- -p[8]
    if (which.max(p[1:3]) == 1) p[1:2] <- p[2:1]
    if (which.max(p[1:3]) == 3) p[2:3] <- p[3:2]
    if (p[5] > p[4]) p[5] = p[4]
    return(p)
  }
  result <- .C("dlogistp",
               as.double(x),
               as.integer(length(x)), 
               as.double(p[1]),
               as.double(p[2]),
               as.double(p[3]),
               as.double(p[4]),
               as.double(p[5]),
               as.double(p[6]),
               as.double(p[7]),
               as.double(p[8]),
               as.integer(1),
               result=double(length(x)))
  return(array(result[["result"]], dim=c(1, length(x))))
}

# dlogistn (dlogist negative, first down, then up):
#    y0 + (y1 - y0) / (1 + ((x1 - x) / xd0) ^ (4 * xd0 * m0 / (y1 - y0))), if x < x1
#    y2 + (y1 - y2) / (1 + ((x - x1) / xd2) ^ (4 * xd2 * m2 / (y2 - y1))), otherwise,
#    constraints: y0, y1, y2, m2, x1, xd0, xd2 >= 0 && m0 <= 0 && xd0 <= x1 && y1 < y0 && y1 < y2
dlogistn <- function(x=NULL, p=NULL) {
  # y0, y1, y2, x1, xd0, xd2, m0, m2
  #  1,  2,  3,  4,   5,   6,  7,  8
  if (is.null(p)) return(8)
  if (is.null(x)) {
    p <- abs(p)
    p[7] <- -p[7]
    if (which.min(p[1:3]) == 1) p[1:2] <- p[2:1]
    if (which.min(p[1:3]) == 3) p[2:3] <- p[3:2]
    if (p[5] > p[4]) p[5] = p[4]
    return(p)
  }
  result <- .C("dlogistn",
               as.double(x),
               as.integer(length(x)), 
               as.double(p[1]),
               as.double(p[2]),
               as.double(p[3]),
               as.double(p[4]),
               as.double(p[5]),
               as.double(p[6]),
               as.double(p[7]),
               as.double(p[8]),
               as.integer(1),
               result=double(length(x)))
  return(array(result[["result"]], dim=c(1, length(x))))
}

fcts <- list(constant, logistp, logistn, dlogistp, dlogistn)



############################################
# function parameter estimators
############################################

estimate.constant.params <- function(data, plot=FALSE, f=0.2) {
  return(max(median(data[, 2]), 0))
}

estimate.logistp.params <- function(data, plot=FALSE, f=0.2) {
  # y0, y2, x1, m1
  l <- lowess(data, f=f)
  n <- length(l$y)

  # note: y0 and y2 can be negative
  y0 <- min(l$y)
  y2 <- max(l$y)
  
  ymid <- (y0 + y2) / 2

  # we look for the middle place where l$y<=ymid, and then ymid<=l$y
  midpoints <- which((l$y[1:(n-1)] <= ymid) & (ymid <= l$y[2:n]))
  # note that in midpoints there can only be numbers between 1 and n-1

  if (length(midpoints) == 0) { # if there is no such place, we take the middle one and m<-0
    x1 <- l$x[round(n/2)]
    m1 <- 0
  } else { # if there is
    mid <- midpoints[ceiling(length(midpoints)/2)]
    x1 <- l$x[mid]
    m1 <- (l$y[mid+1] - l$y[mid]) / (l$x[mid+1]-l$x[mid])
  }

  if (plot) {
    plot(data)
    lines(l)
    
    lines(c(par("usr")[1], l$x[1]), c(y0, y0), col="red")
    lines(c(tail(l$x, 1), par("usr")[2]), c(y2, y2), col="red")
    lines(c(x1, x1), par("usr")[3 : 4], col="red")
    lines(par("usr")[1 : 2], rep((y0 + y2) / 2, 2), col="red")
    
    show.me(function(x) { ((y0 + y2) / 2 - m1 * x1) + m1 * x }, col="red")
    show.me(function(x) { logistp(x, c(y0, y2, x1, m1)) }, col="green")
  }
  
  return(c(y0, y2, x1, m1))
}

estimate.logistn.params <- function(data, plot=FALSE, f=0.2) {
  # y0, y2, x1, m1
  l <- lowess(data, f=f)
  n <- length(l$y)

  # note: y0 and y2 can be negative
  y0 <- min(l$y)
  y2 <- max(l$y)
  
  ymid <- (y0 + y2) / 2

  # we look for the middle place where l$y>=ymid, and then ymid>=l$y
  midpoints <- which((l$y[1:(n-1)] >= ymid) & (ymid >= l$y[2:n]))
  # note that in midpoints there can only be numbers between 1 and n-1

  if (length(midpoints) == 0) { #if there is no such place, we take the middle one and m<-0
    x1 <- l$x[round(n/2)]
    m1 <- 0
  } else { # if there is
    mid <- midpoints[ceiling(length(midpoints)/2)]
    x1 <- l$x[mid]
    m1 <- (l$y[mid+1] - l$y[mid]) / (l$x[mid+1]-l$x[mid])
  }

  if (plot) {
    plot(data)
    lines(l)
    
    lines(c(par("usr")[1], l$x[1]), c(y0, y0), col="red")
    lines(c(tail(l$x, 1), par("usr")[2]), c(y2, y2), col="red")
    lines(c(x1, x1), par("usr")[3 : 4], col="red")
    lines(par("usr")[1 : 2], rep((y0 + y2) / 2, 2), col="red")
    
    show.me(function(x) { ((y0 + y2) / 2 - m1 * x1) + m1 * x }, col="red")
    show.me(function(x) { logistn(x, c(y0, y2, x1, m1)) }, col="green")
  }
  
  return(c(y0, y2, x1, m1))
}

estimate.dlogistp.params <- function(data, plot=FALSE, f=0.2) {
  # parameters:
  # y0:  value at -inf
  # y1:  value at x1
  # y2:  value at +inf
  # x1:  value of reaching y1
  # xd0: at x1-xd the value is (y0+y1)/2
  # xd2: at x1+xd the value is (y1+y2)/2
  # m0:  slope at (x1-xd)
  # m2:  slope at (x1+xd)
  
  l <- lowess(data, f=f)

  #note: y0, y1 and y2 can be negative
  y0 <- head(l$y, 1)
  y2 <- tail(l$y, 1)
  y1 <- max(l$y)

  x1 <- l$x[which.max(l$y)] # which.max always finds the first maximum
  mid0 <- head(which(l$y >= (y0 + y1) / 2), 1)
  mid2 <- tail(which(l$y >= (y1 + y2) / 2), 1)

  if (mid0 == dim(data)[1]) mid0 <- mid0 - 1
  if (mid2 == 1) mid2 <- 2

  xd0 <- x1 - l$x[mid0]
  xd2 <- l$x[mid2] - x1

  m0 <- (l$y[mid0 + 1] - l$y[mid0]) / (l$x[mid0 + 1] - l$x[mid0])
  m2 <- (l$y[mid2] - l$y[mid2 - 1]) / (l$x[mid2] - l$x[mid2 - 1])
  
  if (plot) { 
    plot(data)
    lines(l)
    lines(c(par("usr")[1], l$x[1]), c(y0, y0))
    lines(par("usr")[1 : 2], c(y1, y1))
    lines(c(tail(l$x, 1), par("usr")[2]), c(y2, y2))
    lines(c(par("usr")[1], x1 - xd0), rep((y1 + y0) / 2, 2))
    lines(c(x1 + xd2, par("usr")[2]), rep((y1 + y2) / 2, 2))
    lines(c(x1, x1), par("usr")[3 : 4])
    lines(c(x1 - xd0, x1 - xd0), par("usr")[3 : 4])
    lines(c(x1 + xd2, x1 + xd2), par("usr")[3 : 4])
    show.me(function(x) { dlogistp(x, list(y0=y0, y1=y1, y2=y2, x1=x1, xd0=xd0, xd2=xd2, m0=m0, m2=m2)) }, col="red")
  }

  return(c(y0, y1, y2, x1, xd0, xd2, m0, m2))
}

estimate.dlogistn.params <- function(data, plot=FALSE, f=0.2) {
  # parameters:
  # y0:  value at -inf
  # y1:  value at x1
  # y2:  value at +inf
  # x1:  value of reaching y1
  # xd0: at x1-xd the value is (y0+y1)/2
  # xd2: at x1+xd the value is (y1+y2)/2
  # m0:  slope at (x1-xd)
  # m2:  slope at (x1+xd)

  l <- lowess(data, f=f)

  # note: y0, y1 es y2 can be negative
  y0 <- head(l$y, 1)
  y2 <- tail(l$y, 1)
  y1 <- min(l$y)

  x1 <- l$x[which.min(l$y)] # which.min always finds the first minimum
  mid0 <- head(which(l$y <= (y0 + y1) / 2), 1)
  mid2 <- tail(which(l$y <= (y1 + y2) / 2), 1)

  if (mid0 == dim(data)[1]) mid0 <- mid0 - 1
  if (mid2 == 1) mid2 <- 2

  xd0 <- x1 - l$x[mid0]
  xd2 <- l$x[mid2] - x1

  m0 <- (l$y[mid0 + 1] - l$y[mid0]) / (l$x[mid0 + 1] - l$x[mid0])
  m2 <- (l$y[mid2] - l$y[mid2 - 1]) / (l$x[mid2] - l$x[mid2 - 1])
  
  if (plot) { 
    plot(data)
    lines(l)
    lines(c(par("usr")[1], l$x[1]), c(y0, y0))
    lines(par("usr")[1 : 2], c(y1, y1))
    lines(c(tail(l$x, 1), par("usr")[2]), c(y2, y2))
    lines(c(par("usr")[1], x1 - xd0), rep((y1 + y0) / 2, 2))
    lines(c(x1 + xd2, par("usr")[2]), rep((y1 + y2) / 2, 2))
    lines(c(x1, x1), par("usr")[3 : 4])
    lines(c(x1 - xd0, x1 - xd0), par("usr")[3 : 4])
    lines(c(x1 + xd2, x1 + xd2), par("usr")[3 : 4])
    show.me(function(x) { dlogistn(x, list(y0=y0, y1=y1, y2=y2, x1=x1, xd0=xd0, xd2=xd2, m0=m0, m2=m2)) }, col="red")
  }

  return(c(y0, y1, y2, x1, xd0, xd2, m0, m2))
}

estimators <- list(estimate.constant.params, estimate.logistp.params, estimate.logistn.params, estimate.dlogistp.params, estimate.dlogistn.params)



############################################
# main algorithm (function fitting)
############################################

fit <- function(data, fct, estimator, lambda, fix.starting=FALSE, baseline.part.if.fixed=NA) {
  # can only fit constant, (d)logist[p|n]

  deviation <- function(x) { sum(abs(fct(data[, 1], x) - data[, 2]) ^ lambda) }

  if (fct() == 1) {
    par <- estimator(data)
    sad <- deviation(par)
    return(list(par=par, sad=sad))
  }

  # if we fix parameter y0 at the median value of the first baseline.part.if.fixed seconds
  # (we replace the deviation and estimator functions to use this value)
  if (fix.starting) {
    first.seconds <- which(data[, 1]<=baseline.part.if.fixed)
    if (length(first.seconds) > 0)
      y0 <- median(data[first.seconds, 2])
    else
      y0 <- data[1, 2]
    deviation <- function(x) { sum(abs(fct(data[, 1], c(y0, x)) - data[, 2]) ^ lambda) }

    estimator.original <- estimator
    # we don't need parameter y0 which was made fix
    estimator <- function(data, f=0.2) { return(tail(estimator.original(data=data, f=f), -1)) }

    m.4 <- 3  # number of m  parameter when using a 4-parameter function
    m0.8 <- 6 # number of m0 parameter when using a 8-parameter function
    m2.8 <- 7 # number of m2 parameter when using a 8-parameter function
  } else {
    m.4 <- 4  # number of m  parameter when using a 4-parameter function
    m0.8 <- 7 # number of m0 parameter when using a 8-parameter function
    m2.8 <- 8 # number of m2 parameter when using a 8-parameter function
  }
  
  # 1. NM, f=0.05
  start <- estimator(data, f=0.05)
  par <- optim(start, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
  mindevi <- deviation(par)
  
  # 2-4. NM, f=0.1, 0.15, 0.35 (0.2 kesobb jon majd)
  for (f in c(0.05, 0.1, 0.15, 0.35)) {
    start <- estimator(data, f=f)
    par1 <- optim(start, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
    if (deviation(par1) < mindevi) { par <- par1; mindevi <- deviation(par) }
  }
  
  # 5-9. SANN, f=0.05, 0.1, 0.2, 0.15, 0.35
  for (f in c(0.05, 0.1, 0.2, 0.15, 0.35)) {
    start <- estimator(data, f=f)
    par1 <- optim(start, deviation, control=list(maxit=10000), method="SANN")$par
    if (deviation(par1) < mindevi) { par <- par1; mindevi <- deviation(par) }
  }
  
  # logist:  10-14. NM, f=0.2, we modify the slope parameter
  # dlogist: 10-34. NM, f=0.2, we modify the slope parameters
  start0 <- estimator(data)
  if (fct() == 4) {
    for (m in c(0.01, 0.1, 1, 10, 100)) {
      start[m.4] <- start0[m.4]
      start[m.4] <- start[m.4] * m
      par1 <- optim(start, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
      if (deviation(par1) < mindevi) { par <- par1; mindevi <- deviation(par) }
    }
  } else {
    for (m0 in c(0.01, 0.1, 1, 10, 100)) {
      for (m2 in c(0.01, 0.1, 1, 10, 100)) {
        start[m0.8] <- start0[m0.8] * m0
        start[m2.8] <- start0[m2.8] * m2
        par1 <- optim(start, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
        if (deviation(par1) < mindevi) { par <- par1; mindevi <- deviation(par) }
      }
    }
  }
  #logist:  15-19. final refinement
  #dlogist: 35-39. final refinement
  par <- optim(par, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
  par <- optim(par, deviation, control=list(maxit=10000), method="SANN")$par
  par <- optim(par, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
  par <- optim(par, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
  par <- optim(par, deviation, control=list(maxit=10000), method="Nelder-Mead")$par
  
  # if fix.starting is used we have to make this trick when calling fct(NULL, par)
  if (fix.starting)
    par <- tail(fct(NULL, c(y0, par)), -1)
  else
    par <- fct(NULL, par)
  sad <- deviation(par)

  # if fix.starting is used, we have to put back y0 to the beginning
  if (fix.starting) par <- c(y0, par)
  
  return(list(par=par, sad=sad))
}

cv10fold <- function(data, fct, estimator, lambda, plot=FALSE, fix.starting=FALSE, baseline.part.if.fixed=NA) {
  n <- dim(data)[1] # TODO: it is possible that there are less than 100 time points
  s <- sample(1 : n)
  step <- n / 10
  sad <- 0
  test.sets <- training.sets <- list()
  for (i in 1 : 10) {
    if (i == 1) {           training.set <- data[s[ceiling(1+step) : n], ]
    } else { if (i == 10) { training.set <- data[s[1 : (ceiling(1+step*9) - 1)], ]
    } else {                training.set <- data[s[c(1 : (ceiling(1+step*(i-1)) - 1), (ceiling(1+step*i)) : n)], ] } }
    par <- fit(training.set, fct, estimator, lambda, fix.starting=fix.starting, baseline.part.if.fixed=baseline.part.if.fixed)$par
    test.set <- matrix(data[s[ceiling(1+step*(i-1)) : (ceiling(1+step*i) - 1)], ], ncol=2)
    sad <- sad + sum(abs(fct(test.set[, 1], par) - test.set[, 2]) ^ lambda)

    training.sets[[i]] <- training.set
    test.sets[[i]] <- test.set
    if (plot) {
      plot(training.set, main=i, xlab=sum(abs(fct(test.set[, 1], par) - test.set[, 2]) ^ lambda))
      points(test.set, col="red")
      show.me(function(x) { fct(x, par) })
    }
  }
  return(list(sad=sad, training.sets=training.sets, test.sets=test.sets))
}


############################################
# input and output of kinetics files
############################################

write.kinetics.file <- function(filename, metadata, kinetics.data, medians) {
  # format:                      string    matrix    see below      matrix
  # kinetics.data should be a list, each element of which is a list with the following fields:
  #   *name (string)
  #   *cv.value (1 number)
  #   *sad (1 number)
  #   *par (matrix)

  write.table(1 + 1 + dim(metadata)[1] + 1 + length(kinetics.data) * (1+3+1+dim(kinetics.data[[1]]$par)[1] + 1 + dim(medians)[1]),
                                     filename, append=F, col.names=FALSE, row.names=FALSE, quote=FALSE)
  write.table(dim(metadata)[1],      filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
  write.table(metadata,              filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
  write.table(length(kinetics.data) * (1+3+1+dim(kinetics.data[[1]]$par)[1]),
                                     filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)

  for (fct in kinetics.data) {
    write.table(3,                   filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
    write.table(fct$name,            filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
    write.table(fct$cv.value,        filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
    write.table(fct$sad,             filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
    write.table(dim(fct$par)[1],     filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
    write.table(fct$par,             filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
  }
  write.table(dim(medians)[1],       filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
  write.table(medians,               filename, append=T, col.names=FALSE, row.names=FALSE, quote=FALSE)
}

read.kinetics.file <- function(filename) {
  nall <- as.numeric(read.table(filename, skip=0, nrows=1))
  nmeta <- as.numeric(read.table(filename, skip=1, nrows=1))
  if (nmeta > 0) {
    metadata <- read.table(filename, skip=2, nrows=nmeta, sep="\n", blank.lines.skip=FALSE)
  } else {
    metadata <- data.frame()
  }
  nkin <- as.numeric(read.table(filename, skip=(2 + nmeta), nrows=1))
  n <- 0
  kinetics.data <- list()
  num <- 0
  while (n < nkin) { #ha = lesz vele, mar beolvastuk
    nkinmeta <-                  as.numeric(read.table(filename, skip=(2 + nmeta + 1 + n), nrows=1))
    kin.data <- list()
    kin.data$name <- as.character(as.matrix(read.table(filename, skip=(2 + nmeta + 1 + n + 1), nrows=1)))

    kin.data$cv.value <-         as.numeric(read.table(filename, skip=(2 + nmeta + 1 + n + 2), nrows=1))
    kin.data$sad      <-         as.numeric(read.table(filename, skip=(2 + nmeta + 1 + n + 3), nrows=1))
    if (nkinmeta > 3) nothing <-            read.table(filename, skip=(2 + nmeta + 1 + n + 4), nrows=(nkinmeta-3))
    npar <-                      as.numeric(read.table(filename, skip=(2 + nmeta + 1 + n + 1 + nkinmeta), nrows=1))
    kin.data$par <-               as.matrix(read.table(filename, skip=(2 + nmeta + 1 + n + 1 + nkinmeta + 1), nrows=npar))
    colnames(kin.data$par) <- NULL
    rownames(kin.data$par) <- NULL
    n <- n + nkinmeta + npar + 2
    num <- num + 1
    kinetics.data[[num]] <- kin.data
  }
  nmedian <- as.numeric(read.table(filename, skip=(2 + nmeta + 1 + n), nrows=1))
  medians <- as.matrix(read.table(filename, skip=(2 + nmeta + 1 + n + 1), nrows=nmedian))
  colnames(medians) <- NULL
  rownames(medians) <- NULL
  return(list(metadata=metadata, kinetics.data=kinetics.data, medians=medians))
}



############################################
# putting it all together
############################################

# when using fix.starting, caflux.analysis appends a row "fix.starting.seconds=..." to the metadata contained in filename.in
caflux.analysis <- function(filename.in, filename.out) {
  a <- read.gated.file(filename.in)

  if (length(a$metadata) > 6) {
    n <- round(as.numeric(a$metadata[7]))
    if (n < 1 || n > 201)
      n <- 201
  } else
    n <- 1

  len0 <- dim(a$data)[1]
  a$data <- matrix(as.numeric(a$data), ncol=2)
  a$data <- a$data[which(!is.na(a$data[, 1])), ]
  a$data <- a$data[which(!is.na(a$data[, 2])), ]
  len1 <- dim(a$data)[1]
  if (len1/len0 < 0.5)
    stop("Too many NAs in input data (ie more than 50%).")
  a$data[, 1] <- retime(a$data[, 1])
  a$data[, 1] <- dividetime(a$data[, 1], 100) #max. 100 idopillanat van!
  quants <- quantiles(n)
  qdata <- quantiles.from.data(a$data, quants)
  times <- unique(a$data[, 1])
  metadata <- a$metadata
  rm(a)
  q1data <- cbind(times, qdata[round(quantile(1:n, 0.25)), ])
  q2data <- cbind(times, qdata[round(quantile(1:n, 0.50)), ])
  q3data <- cbind(times, qdata[round(quantile(1:n, 0.75)), ])

  if (length(times) < 50)
    stop("Too many gaps in the experiment.")

  if (length(metadata) > 7) {
    baseline.part.if.fixed <- as.numeric(metadata[8])
    fix.starting <- baseline.part.if.fixed >= times[1]
  } else {
    fix.starting <- FALSE
    baseline.part.if.fixed <- NA
  }

  kinetics.data <- list()
  for (num in 1 : length(fcts)) {
    fct <- fcts[[num]]
    estimator <- estimators[[num]]
    cat("fct:", fct())
    cv.value <- cv10fold(q1data, fct, estimator, lambda, fix.starting=fix.starting)$sad + cv10fold(q2data, fct, estimator, lambda, fix.starting=fix.starting)$sad + cv10fold(q3data, fct, estimator, lambda, fix.starting=fix.starting, baseline.part.if.fixed=baseline.part.if.fixed)$sad
    cat(" -> ")
    pars <- array(NA, dim=c(n, fct()))
    sad <- 0
    for (i in 1 : n) {
      model <- fit(cbind(times, qdata[i, ]), fct, estimator, lambda, fix.starting=fix.starting, baseline.part.if.fixed=baseline.part.if.fixed)
      pars[i, ] <- model$par
      sad <- sad + model$sad
      cat(i); cat(" ")
    }
    cat("\n")
    kinetics.data[[num]] <- list(name=fct.names[[num]], cv.value=cv.value, sad=sad, par=pars)
  }

  if (fix.starting)
    metadata <- rbind(metadata, paste("fix.starting.seconds=", baseline.part.if.fixed, sep=""))
  write.kinetics.file(filename.out, metadata, kinetics.data, q2data)
}
